#!/usr/bin/python

# stdlib:
import os, sys
import base64
import xmlrpclib
import re
import time
from xmlrpclib import Fault

# otherlib:
from rpcproxy import Connection, Database, Object
import argparse

__version__ = "0.1"


class Prog(object):
    def __init__(self, args):
        self.args = args
        self.verbose = self._get_verbosity()

        self.db = self.create_db_connection()
        self.cnx = self.connect()

    def _get_verbosity(self):
        return self.args.verbose and sum(self.args.verbose) or 0

    def create_db_connection(self):
        db = Database()
        return db

    def _connect(self):
        if os.path.isfile(os.path.expanduser(self.args.psw)):
            psw = base64.decodestring(open(os.path.expanduser(self.args.psw), 'r').read().replace('\n', ''))
        else:
            psw = self.args.psw

        return Connection(
            server=self.args.host,
            port=self.args.port,
            dbname=self.args.database,
            login=self.args.user,
            password=psw,
            protocol=self.args.protocol)

    def connect(self):
        try:
            return self._connect()
        except Exception, e:
            sys.stderr.write("Error: unable to connect. Reason: %s\n" % (e, ))
            sys.stderr.flush()
            raise

    def run(self):
        if self.args.cmd == 'list':
            model_obj = Object(self.cnx, "ir.model")
            model_ids = model_obj.search([])
            models = model_obj.read(model_ids, ['model'])
            lst = sorted([m['model'] for m in models])
            sys.stdout.write('\n'.join(lst))
            sys.stdout.flush()
        elif self.args.cmd.startswith('print '):
            model = self.args.cmd.replace('print ', '')
            obj = Object(self.cnx, model)
            limit = self.args.limit and int(self.args.limit) or False
            ids = obj.search([], 0, limit)
            if self.args.fields:
                fields = [f.strip() for f in self.args.fields.split(',')]
            else:
                fields = []
            objs = obj.read(ids, fields)
            from pprint import pprint
            pprint(objs)
            sys.stdout.write("fields: %s\n" % (fields, ))
            sys.stdout.write("len: %s\n" % (len(objs), ))
            sys.stdout.flush()
        elif self.args.cmd == 'all':
            if self.args.exclude_modules:
                excluded = self.args.exclude_modules.replace(' ', '').split(',')
            else:
                excluded = []

            model_obj = Object(self.cnx, "ir.model")
            model_ids = model_obj.search([])
            models = [rec['model'] for rec in model_obj.read(model_ids, ['model'])]
            models.sort()

            for model in models:
                if model in excluded:
                    continue

                if self.args.show_model_before_process:
                    sys.stdout.write("%s,xxx\n" % (model, ))
                    sys.stdout.flush()

                obj = Object(self.cnx, model)
                limit = self.args.limit and int(self.args.limit) or False
                try:
                    ids = obj.search([], 0, limit)
                except (Exception, Fault), e:
                    sys.stderr.write("Error in model '%s': %s\n" % (model, e))
                    sys.stderr.flush()
                    continue
                except (KeyboardInterrupt, RuntimeError, SystemExit), e:
                    raise
                except:
                    sys.stderr.write("Uncaught error in model '%s'\n" % (model, ))
                    sys.stderr.flush()
                    continue

                if self.args.fields:
                    fields = [f.strip() for f in self.args.fields.split(',')]
                else:
                    fields = []
                try:
                    objs = obj.read(ids, fields)
                except (Exception, Fault), e:
                    sys.stderr.write("Error in model '%s': %s\n" % (model, e))
                    sys.stderr.flush()
                    continue
                except:
                    sys.stderr.write("Uncaught error in model '%s'\n" % (model, ))
                    sys.stderr.flush()
                    continue
                if self.args.server_log:
                    last_lines = tail(self.args.server_log, 20)

                    sql_sum = '??'
                    for line in last_lines:
                        if line.find("SUM into") > -1:
                            m = re.search(r'SUM into.*/([0-9]*)', line)
                            if m:
                                sql_sum = line[m.start():m.end()].split('/')[-1]
                else:
                    sql_sum = '?'
                sys.stdout.write("%s,%s,%s\n" % (model, len(objs), sql_sum))
                sys.stdout.flush()
        else:
            raise Exception("Error: unknown command")

def tail(fn, window=20 ):
    f = open(fn, 'r')
    f.seek( 0, 2 )
    bytes= f.tell()
    size= window
    block= -1
    while size > 0 and bytes+block*1024  > 0:
        # If your OS is rude about small files, you need this check
        # If your OS does 'the right thing' then just f.seek( block*1024, 2 )
        # is sufficient
        if (bytes+block*1024 > 0):
            ##Seek back once more, if possible
            f.seek( block*1024, 2 )
        else:
            #Seek to the beginning
            f.seek(0, 0)
        data= f.read( 1024 )
        linesFound= data.count('\n')
        size -= linesFound
        block -= 1
    f.seek( block*1024, 2 )
    f.readline() # find a newline
    lastBlocks= list( f.readlines() )
    f.close()
    return lastBlocks[-window:]


def _main():
    parser = argparse.ArgumentParser(prefix_chars='-+', add_help=False, conflict_handler='resolve')

    # options:
    g2 = parser.add_argument_group("optional arguments")
    g2.add_argument('--host',
                    help="host. Default: %(default)s",
                    dest="host", action="store", default='localhost')
    g2.add_argument('-p', '--port', type=int,
                    help="xml-rpc port. Default: %(default)s",
                    dest="port", action="store", default='8069')
    g2.add_argument('--user',
                    help="admin login. Default: %(default)s",
                    dest="user", action="store", default='admin')
    g2.add_argument('--password',
                    help="admin password. Default: %(default)s",
                    dest="psw", action="store", default='admin')
    g2.add_argument('--protocol',
                    help="protocol. Default: %(default)s",
                    dest="protocol", action="store", default='http')
    g2.add_argument('-v', '--verbose',
                    help="Verbose output.. Default: %(default)s", dest="verbose",
                    action='append_const', const=1)
    g2.add_argument('-l', '--limit',
                    help="Limit. Default: %(default)s",
                    dest="limit", action='store', default=False)
    g2.add_argument('-f', '--fields',
                    help="Fields. Default: all",
                    dest="fields", action='store', default=False)
    g2.add_argument('--server-log',
                    help="Server log file. Default: %(default)s",
                    dest="server_log", action='store', default=False)
    g2.add_argument('--exclude-modules',
                    help="Server log file. Default: %(default)s",
                    dest="exclude_modules", action='store', default=False)
    g2.add_argument('--show-model-before-process',
                    help="Output the model name just before processing it (for debugging purpose). Default: %(default)s",
                    dest="show_model_before_process", action='store_true', default=False)



    # help and version:
    g2.add_argument('-V', '--version',
                    help="Show this program version and exit.", dest="vers",
                    action='store_true')
    g2.add_argument('-h', '--help',
                    help="Show this help message and exit.", dest="help",
                    action='store_true')

    g1 = parser.add_argument_group("required arguments")
    # dummy add required args without the required attrs for 'help' message:
    g1.add_argument('+db', metavar="DATABASE",
                    help="Database (required)", dest='database',
                   )
    g1.add_argument('+cmd', metavar="CMD",
                    help="command (required)", dest='cmd',
                   )

    args = parser.parse_args()

    if args.help:
        parser.print_help()
    elif args.vers:
        sys.stdout.write(parser.prog+' '+__version__+'\n')
        sys.stdout.flush()
    else:
        # required args:
        g1.add_argument('+db', metavar="DATABASE", required=True,
                         help="Database (required)", dest='database',
                        )
        g1.add_argument('+cmd', metavar="CMD", required=True,
                    help="command (required)", dest='cmd',
                   )
        args = parser.parse_args()
        prog = Prog(args)
        prog.run()

if __name__ == '__main__':
    _main()

